import datetime

# property type collective.properties form knows how to handle
PROPERTY_TYPES = ('boolean', 'date', 'float', 'int', 'lines', 'long', 'string',
    'text', 'tokens', 'selection', 'multiple selection')
    
# textarea widget dimensions for text property types on manage properties form
TEXT_FIELD_WIDGET_DIMENSIONS = (60, 10)

# default datetime for date property type field
DEFAULT_DATE_VALUE = datetime.datetime(1984, 06, 17)



import types
from datetime import datetime

from DateTime import DateTime

from zope.interface import Interface, implements, Attribute
from zope import schema

from collective.properties import messageFactory as _
from collective.properties.config import DEFAULT_DATE_VALUE
from collective.properties.field import BytesLineTuple


# keeps mapping of IPropertyManager property types to
# interfaces for z3c.form forms to display it in CRUD listing
PROPERTY_MAP = {}

def registerPropertyType(ptype, prop):
    assert ptype not in PROPERTY_MAP
    PROPERTY_MAP[ptype] = prop

def unregisterPropertyType(ptype):
    assert ptype in PROPERTY_MAP
    del PROPERTY_MAP[ptype]

# abstract property to inherit from for all IPropertyManager property types

class IProperty(Interface):
    
    default = Attribute(_(u"Default valid property value."))
    
    id = schema.ASCIILine(
        title=_(u"Property Id"),
        description=u'',
        required=True,
        readonly=True)
    
    ptype = schema.ASCIILine(
        title=_(u"Property Type"),
        description=u'',
        required=True,
        readonly=True)

class Property(object):
    
    implements(IProperty)
    
    default = None
    
    def __init__(self, context, id, value=default, ptype=None,
                 select_variable=None, mode='wd'):
        self.context = context
        self.id = id
        # we don't need to update context's property on property init
        self.__dict__['value'] = value
        self.ptype = ptype
        self.select_variable = select_variable
        self.mode = mode
    
    def _get_value(self):
        return self.__dict__['value']
    
    def _set_value(self, value):
        # on explicit attribute set (property.value = value) we update
        # context's property value
        # Note: we never set value to None, instead we always set it to
        # default property type value, otherwise old manage_propertiesForm
        # breaks, so we're just following the same old rules ;)
        if value is None:
            value = self.default
        
        # check if property is writable
        self._validate()
        
        self.context._updateProperty(self.id, value)
        self.__dict__['value'] = value
    
    value = property(_get_value, _set_value)
    
    def _validate(self):
        """Checks if property is writable"""
        propdict = self.context.propdict()
        if 'w' not in propdict[self.id].get('mode', 'wd'):
            raise schema.ValidationError(u"The property '%s' is not writable." %
                self.id)
    
    def __repr__(self):                                                                                                              
        return "<Property with id=%r, type=%r, value=%r>" % (self.id,
            self.ptype, self.value)

# IPropertyManager 'string' property

class IStringProperty(IProperty):
    value = schema.BytesLine(
        title=_(u"Property Value"),
        description=_(u"Contains text line, including any non-ascii"
                      " characters."),
        required=False,
        default='')

class StringProperty(Property):
    implements(IStringProperty)
    default = ''

registerPropertyType('string', StringProperty)

# IPropertyManager 'int' property

class IIntProperty(IProperty):
    value = schema.Int(
        title=_(u"Property Value"),
        description=_(u"Contains integer value, e.g. -1, 0, 1, 2."),
        required=False,
        default=0)

class IntProperty(Property):
    implements(IIntProperty)
    default = 0

registerPropertyType('int', IntProperty)
registerPropertyType('long', IntProperty)

# IPropertyManager 'float' property

class IFloatProperty(IProperty):
    value = schema.Float(
        title=_(u"Property Value"),
        description=_(u"Contains floating value, e.g. -1.1, 0.0, 1.25."),
        required=False,
        default=0.0)

class FloatProperty(Property):
    implements(IFloatProperty)
    default = 0.0

registerPropertyType('float', FloatProperty)

# IPropertyManager 'lines' property

class ILinesProperty(IProperty):    
    # we keep data in bytes lines to be compliant with OFS.PropertyManager
    value = BytesLineTuple(
        title=_(u"Property Value"),
        description=_(u"Contains list of strings."),
        required=False,
        value_type=schema.BytesLine(),
        default=())

class LinesProperty(Property):
    implements(ILinesProperty)
    default = ()

registerPropertyType('lines', LinesProperty)

# IPropertyManager 'text' property

class ITextProperty(IProperty):
    value = schema.Bytes(
        title=_(u"Property Value"),
        description=_(u"Contains text."),
        required=False,
        default='')

class TextProperty(Property):
    implements(ITextProperty)
    default = ''

registerPropertyType('text', TextProperty)

# IPropertyManager 'boolean' property

class IBooleanProperty(IProperty):
    value = schema.Bool(
        title=_(u"Property Value"),
        description=_(u"Checked for True, otherwise - False."),
        required=False,
        default=False)

class BooleanProperty(Property):
    implements(IBooleanProperty)
    default = False

registerPropertyType('boolean', BooleanProperty)

# IPropertyManager 'date' property

class IDateProperty(IProperty):
    value = schema.Datetime(
        title=_(u"Property Value"),
        description=_(u"Contains date and time value."),
        required=False,
        default=DEFAULT_DATE_VALUE)

class DateProperty(Property):
    implements(IDateProperty)
    default = DEFAULT_DATE_VALUE

    def _get_value(self):
        """Override this method here to convert value from Zope2 DateTime
        to python datetime object.
        """
        # convert DateTime to datetime
        value = self.__dict__['value']
        if isinstance(value, DateTime):
            parts = value.parts()
            value = datetime(*(parts[:5]+(int(parts[5]),)))
        return value
    
    def _set_value(self, value):
        """Override this method here to convert python datetime object
        we got from z3c.form based form to Zope2 DateTime to save into
        date property type.
        """
        # convert python datetime object to Zope2 DateTime
        if isinstance(value, datetime):
            value = DateTime(value.strftime('%Y/%m/%d %H:%M:%S'))
        
        # on explicit attribute set (property.value = value) we update
        # context's property value
        # Note: we never set value to None, instead we always set it to
        # default property type value, otherwise old manage_propertiesForm
        # breaks, so we're just following the same old rules ;)
        if value is None:
            value = self.default
        
        # check if property is writable
        self._validate()
        
        self.context._updateProperty(self.id, value)
        self.__dict__['value'] = value

    value = property(_get_value, _set_value)

registerPropertyType('date', DateProperty)

# IPropertyManager 'tokens' property

class ITokensProperty(IProperty):
    value = schema.BytesLine(
        title=_(u"Property Value"),
        description=_(u"Contains words separated by spaces."),
        required=False,
        default='')

class TokensProperty(Property):
    implements(ITokensProperty)
    default = ''

    def __init__(self, context, id, value=default, ptype=None,
                 select_variable=None, mode='wd'):
        """Override it here to convert value (which is list) to string.
        This is required because IPropertyManger expects string to be
        passed to it while saving tokens property value as list.
        So we decided to use string field and widget to manage tokens
        property type.
        """
        self.context = context
        self.id = id
        self.ptype = ptype
        self.mode = mode
        
        # here we get tuple of tokens, so convert it to string before
        # returning it to string widget
        if isinstance(value, (types.TupleType, types.ListType)):
            value = ' '.join(value)
        
        self.__dict__['value'] = value

registerPropertyType('tokens', TokensProperty)

# IPropertyManager 'selection' property

class ISelectionProperty(IProperty):
    value = schema.Choice(
        title=_(u"Property Value"),
        description=_(u"Contains utf-8 encoded string."),
        required=False,
        vocabulary='collective.properties.SelectionPropertyTypeVocabulary',
        default='',
        missing_value='')

class SelectionProperty(Property):
    implements(ISelectionProperty)
    default = ''

    def _get_value(self):
        """Override this method here to convert string to unicode value
        before returning it to outside world. We can't do this only
        inside data converter as this breaks while z3c.form form compares
        old and new value to check if applying changes is needed.
        """
        # convert utf-8 encoded string to unicode
        value = self.__dict__['value']
        if isinstance(value, str):
            value = value.decode('utf-8')
        return value
    
    def _set_value(self, value):
        """Override this method here to convert value to utf-8
        encoded string. That's how IPropertyManager sets selection
        property type value.
        """
        if value is None:
            value = self.default
        
        # encode unicode value to utf-8 encoded string
        if isinstance(value, unicode):
            value = value.encode('utf-8')
        
        # check if property is writable
        self._validate()
        
        self.context._updateProperty(self.id, value)
        self.__dict__['value'] = value

    value = property(_get_value, _set_value)

registerPropertyType('selection', SelectionProperty)

# IPropertyManager 'multiple selection' property

class IMultipleSelectionProperty(IProperty):
    value = BytesLineTuple(
        title=_(u"Property Value"),
        description=_(u"Contains list of utf-8 encoded strings."),
        required=False,
        value_type=schema.Choice(
            vocabulary='collective.properties.SelectionPropertyTypeVocabulary'),
        default=())

class MultipleSelectionProperty(Property):
    implements(IMultipleSelectionProperty)
    default = ()

    def _get_value(self):
        """Convert list of strings to list of unicodes."""
        # convert utf-8 encoded string to unicode
        value = self.__dict__['value']
        result = []
        for item in value:
            if isinstance(item, str):
                item = item.decode('utf-8')
            result.append(item)
        return tuple(result)
    
    def _set_value(self, value):
        """Convert list of unicodes to list of strings."""
        if value is None:
            value = self.default
        
        # encode unicode value to utf-8 encoded string
        result = []
        for item in value:
            if isinstance(item, unicode):
                item = item.encode('utf-8')
            result.append(item)
        value = tuple(result)

        # check if property is writable
        self._validate()

        self.context._updateProperty(self.id, value)
        self.__dict__['value'] = value

    value = property(_get_value, _set_value)

registerPropertyType('multiple selection', MultipleSelectionProperty)



from zope.interface import alsoProvides, implementsOnly
from zope.schema.interfaces import ITitledTokenizedTerm
from zope.i18n import translate

from z3c.form.widget import FieldWidget
from z3c.form.browser.select import SelectWidget
from z3c.form.browser.orderedselect import OrderedSelectWidget
from z3c.form.browser.widget import HTMLSelectWidget, addFieldClass
from z3c.form.widget import SequenceWidget
from z3c.form.interfaces import NOVALUE

from collective.properties.interfaces import ISpecialSelectWidget, \
    ISpecialOrderedSelectWidget

class SpecialSelectWidget(SelectWidget):
    """Custom Select widget implementation.
    
    What's customized:
        * separate converter which won't break if selected value is not
          within widget terms at the moment
        * widget doesn't use term tokens only term values, we need this fix
          to make IPropertyManager selection property type work with terms
          containing non-ascii characters (token couldn't not contain it) as
          native property manager edit form support it
    """
    
    implementsOnly(ISpecialSelectWidget)

    def isSelected(self, term):
        return term.value in self.value

    def update(self):
        """See z3c.form.interfaces.IWidget."""
        HTMLSelectWidget.update(self)
        SequenceWidget.update(self)
        addFieldClass(self)
        self.items = []
        if (not self.required or self.prompt) and self.multiple is None:
            message = self.noValueMessage
            if self.prompt:
                message = self.promptMessage
            self.items.append({
                'id': self.id + '-novalue',
                'value': self.noValueToken,
                'content': message,
                'selected': self.value == []
                })
        for count, term in enumerate(self.terms):
            selected = self.isSelected(term)
            id = '%s-%i' % (self.id, count)
            content = term.value
            if ITitledTokenizedTerm.providedBy(term):
                content = translate(
                    term.title, context=self.request, default=term.title)
            self.items.append(
                {'id':id, 'value':term.value, 'content':content,
                 'selected':selected})

    @property
    def displayValue(self):
        value = []
        for token in self.value:
            # Ignore no value entries. They are in the request only.
            if token == self.noValueToken:
                continue
            term = self.terms.getTermByValue(token)
            if ITitledTokenizedTerm.providedBy(term):
                value.append(translate(
                    term.title, context=self.request, default=term.title))
            else:
                value.append(term.value)
        return value

    def extract(self, default=NOVALUE):
        """See z3c.form.interfaces.IWidget."""
        if (self.name not in self.request and
            self.name+'-empty-marker' in self.request):
            return []
        value = self.request.get(self.name, default)
        if value != default:
            for token in value:
                if token == self.noValueToken:
                    continue
                try:
                    self.terms.getTerm(token)
                except LookupError:
                    return default
        return value

def SpecialSelectFieldWidget(field, request):
    """IFieldWidget factory for SelectWidget which won't break
    when selected value is not within widget terms.
    And also which doesn't use term tokens but only term values.
    """
    return FieldWidget(field, SpecialSelectWidget(request))

class SpecialOrderedSelectWidget(OrderedSelectWidget):
    """Custom Ordered-Select widget implementation.
    
    We customized it here for the same purposes we did with SelectWidget
    defined above: token to value conversion and assign custom converter.
    """
    implementsOnly(ISpecialOrderedSelectWidget)

    def getItem(self, term, count=0):
        id = '%s-%i' % (self.id, count)
        content = term.value
        if ITitledTokenizedTerm.providedBy(term):
            content = translate(
                term.title, context=self.request, default=term.title)
        return {'id':id, 'value':term.value, 'content':content}

    def update(self):
        """See z3c.form.interfaces.IWidget."""
        HTMLSelectWidget.update(self)
        SequenceWidget.update(self)
        addFieldClass(self)
        self.items = [
            self.getItem(term, count)
            for count, term in enumerate(self.terms)]
        self.selectedItems = [
            self.getItem(self.terms.getTerm(token), count)
            for count, token in enumerate(self.value)]
        self.notselectedItems = self.deselect()

    @property
    def displayValue(self):
        value = []
        for token in self.value:
            # Ignore no value entries. They are in the request only.
            if token == self.noValueToken:
                continue
            term = self.terms.getTermByValue(token)
            if ITitledTokenizedTerm.providedBy(term):
                value.append(translate(
                    term.title, context=self.request, default=term.title))
            else:
                value.append(term.value)
        return value

    def extract(self, default=NOVALUE):
        """See z3c.form.interfaces.IWidget."""
        if (self.name not in self.request and
            self.name+'-empty-marker' in self.request):
            return []
        value = self.request.get(self.name, default)
        if value != default:
            for token in value:
                if token == self.noValueToken:
                    continue
                try:
                    self.terms.getTerm(token)
                except LookupError:
                    return default
        return value

def SpecialSequenceSelectFieldWidget(field, request):
    return FieldWidget(field, SpecialOrderedSelectWidget(request))



from OFS.interfaces import IPropertyManager
from zExceptions import BadRequest

from zope.interface import implementedBy, Interface, Invalid
from zope import schema
from zope.schema.interfaces import IBytesLine, IBytes, IChoice
from zope.component import getMultiAdapter
from zope.event import notify
from zope.lifecycleevent import ObjectCreatedEvent

from Products.Five.browser import BrowserView
from Products.CMFPlone.utils import safe_unicode

from z3c.form import field, button, form
from z3c.form.interfaces import DISPLAY_MODE, ITextAreaWidget, IErrorViewSnippet
from z3c.form.browser.textarea import TextAreaFieldWidget
from plone.z3cform.crud import crud
from plone.z3cform.layout import wrap_form

from collective.properties import messageFactory as _
from collective.properties.config import PROPERTY_TYPES, \
    TEXT_FIELD_WIDGET_DIMENSIONS
from collective.properties.propertytypes import PROPERTY_MAP, IProperty
from collective.properties.widget import SpecialSelectFieldWidget, \
    SpecialSequenceSelectFieldWidget
from collective.properties.field import IBytesLineTuple


class IsPropertyManagerView(BrowserView):
    """Checks if context object provides IPropertyManager interface"""
    
    def __call__(self):
        return IPropertyManager.providedBy(self.context)

class IAddProperty(Interface):

    id = schema.ASCIILine(
        title=_(u"Property Id"),
        description=u'',
        required=True)

    ptype = schema.Choice(
        title=_(u"Property Type"),
        description=u'',
        required=True,
        values=PROPERTY_TYPES,
        default='string')
    
    select_variable = schema.ASCIILine(
        title=_(u"Select Variable"),
        description=_(u"This field is used only for selection and multiple "
                      "selection property types. It provides the name of a "
                      "property or method which returns a list of strings from "
                      "which the selection can be chosen."),
        required=False,
        default='')

class EditSubForm(crud.EditSubForm):
    """Override this form to get different form for every property type
    and to have some widget a bit prettier.
    """
    
    @property
    def fields(self):
        # here we override update schema based on property type
        fields = field.Fields(self._select_field())

        crud_form = self.context.context
        
        # if we're aware of this property type then render it's
        # dedicated update form, otherwise render base property
        # update form w/o value field
        ptype = self.getContent().ptype
        if ptype in PROPERTY_MAP:
            update_schema = implementedBy(PROPERTY_MAP[ptype]).interfaces(
                ).next()
        else:
            update_schema = crud_form.update_schema
        
        if update_schema is not None:
            fields += field.Fields(update_schema)

        view_schema = crud_form.view_schema
        if view_schema is not None:
            view_fields = field.Fields(view_schema)
            for f in view_fields.values():
                f.mode = DISPLAY_MODE
                # This is to allow a field to appear in both view
                # and edit mode at the same time:
                if not f.__name__.startswith('view_'):
                    f.__name__ = 'view_' + f.__name__
            fields += view_fields

        # assign custom widget factories
        for fid, f in fields.items():
            # assign textarea widget to IBytes field, we deliberately
            # keep all string in bytes instead of unicode to be compliant
            # with old good IPropertyManager API
            if IBytes.providedBy(f.field) and \
               not IBytesLine.providedBy(f.field):
                f.widgetFactory = TextAreaFieldWidget
            
            # assign SpecialSelectWidget to Choice fields to prevent breakage on
            # values that are not within current widget terms
            if IChoice.providedBy(f.field):
                f.widgetFactory = SpecialSelectFieldWidget
            
            # assign SpecialSequenceSelectFieldWidget to IBytesLineTuple fields
            # with IChoice field as it's value_type
            if IBytesLineTuple.providedBy(f.field) and \
               IChoice.providedBy(f.field.value_type):
                f.widgetFactory = SpecialSequenceSelectFieldWidget

        return fields

    def updateWidgets(self):
        super(EditSubForm, self).updateWidgets()
        # customize field widgets a bit
        for wid, widget in self.widgets.items():
            # enlarge all text field widgets
            if ITextAreaWidget.providedBy(widget):
                widget.cols, widget.rows = TEXT_FIELD_WIDGET_DIMENSIONS
        
        # disable 'delete' checkbox for non-deletable properties
        mode = self.getContent().mode
        if 'd' not in mode:
            # deselect and disable checkbox
            select = self.widgets['select']
            select.value = ()
            select.disabled = 'disabled'
            # need to update checkbox widget in order to get updated value
            select.update()
        
        # 'value' field set to readonly mode for non-writable properties
        if 'w' not in mode:
            # set to readonly mode
            value = self.widgets['value']
            value.readonly = 'readonly'

class EditForm(crud.EditForm):
    editsubform_factory = EditSubForm

class AddForm(crud.AddForm):

    @button.buttonAndHandler(_('Add'), name='add')
    def handle_add(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = form.AddForm.formErrorsMessage
            return
        try:
            item = self.context.add(data)
        except schema.ValidationError, e:
            # notify form about error so that status message will be
            # displayed in red and keep previously entered field values
            error = getMultiAdapter((Invalid(e), self.request, None,
                None, self, self.context.context), IErrorViewSnippet)
            error.update()
            self.widgets.errors += (error,)
            self.status = form.AddForm.formErrorsMessage
        else:
            notify(ObjectCreatedEvent(item))
            self.status = _(u"Item added successfully.")

class ManagePropertiesForm(crud.CrudForm):
    """Properties management view"""
    
    label = _(u"Manage Properties Form")
    description = _(u"On this form you can add, update and delete "
                    "IPropertyManager related properties for this context "
                    "object. To add property use Adding form at the bottom of "
                    "this page. Firstly fill in property id and pick up it's "
                    "type and only after it's added you'll be able to fill it's"
                    " value in. For selection property types you'll also need "
                    " to enter it's select variable (it provides the name of a "
                    "property or method which returns a list of strings from "
                    "which the selection(s) can be chosen). Some properties "
                    "could be protected so that you won't be able to remove or "
                    "update it. For non-deletable properties 'select' checkbox "
                    "is disabled, and for non-writable ones 'Property Value' "
                    "input is in 'readonly' mode. For more info on property "
                    "types you can manage here, please, refer to "
                    "collective.properties package README.txt.")
    
    # default update_schema for those property types we're not aware of;
    # actually not used, instead we skip any property types we don't
    # have appropriate handlers for
    update_schema = IProperty
    
    # add schema is the same for all property types, we don't include
    # value field here and adding all property types with default predefined
    # values, so that user has to update it after addition
    add_schema = IAddProperty
    
    # we override it here in order to have different update schemas for
    # different property types
    editform_factory = EditForm
    
    # override add form to display validation problem as error message
    # during addition; currently standard CRUD form displays it as info
    # message
    addform_factory = AddForm
    
    def get_items(self):
        """Return list of properties assigned to context object."""
        props = []
        context = self.context
        get = context.getPropertyType
        propsdict = context.propdict()
        for pid, pvalue in context.propertyItems():
            ptype = get(pid)
            
            # skip any non-defined property types
            if ptype not in PROPERTY_MAP:
                continue
            
            props.append((pid,
                PROPERTY_MAP[ptype](context, pid, pvalue, ptype,
                    propsdict[pid].get('select_variable'),
                    propsdict[pid].get('mode', 'wd'))))
        
        return tuple(props)
    
    def add(self, data):
        """Adds new property with default value."""
        ptype = data['ptype']
        
        # prepare initial property value based on it's type
        if ptype in ('selection', 'multiple selection'):
            # for selection types initial value will be set by PropertyManager
            # and in a value we should pass to it select_variable to get list
            # of available value options to choose from
            value = data['select_variable']
            if value is None:
                value = ''
        elif ptype not in PROPERTY_MAP:
            # if we don't know how to handle this property type then try adding
            # it with empty string as default initial value
            value = ''
        else:
            # get default value
            value = PROPERTY_MAP[ptype].default
        
        # catch all errors during property addition and display it in
        # user-friendly manner on form as invalid message
        try:
            self.context.manage_addProperty(data['id'], value, ptype)
        except Exception, e:
            raise schema.ValidationError(safe_unicode(e.args[0]))
    
    def remove(self, (id, item)):
        """Removes existing property from context object."""
        context = self.context
        if 'd' not in item.mode or id in context._reserved_names:
            raise schema.ValidationError(u"The property '%s' can not be "
                "deleted." % id)
        context._delProperty(id)

ManagePropertiesView = wrap_form(ManagePropertiesForm)



from zope.schema import Tuple
from zope.interface import implements

from collective.properties.interfaces import IBytesLineTuple


class BytesLineTuple(Tuple):
    """See interface"""
    implements(IBytesLineTuple)



from zope.schema.interfaces import ITuple
from z3c.form.interfaces import ISelectWidget, IOrderedSelectWidget


class IBytesLineTuple(ITuple):
    """Tuple containing byte strings inside.
    We use separate field to attach custom data converter
    from byte string to text area widget.
    """

class ISpecialSelectWidget(ISelectWidget):
    """SelectWidget which doesn't break if there is not current value
    inside widget terms.
    Also it handles any data types correctly and inline with how
    IPropertyManager does it, so /manage_properitesForm still works.
    """

class ISpecialOrderedSelectWidget(IOrderedSelectWidget):
    """Ordered Select Widget which doesn't break if current field value contains
    items which are not within widget terms.
    Also it handles any data types correctly and inline with how
    IPropertyManager does it, so /manage_properitesForm still works.
    """



from setuptools import setup, find_packages
import os

version = '1.0a2'

setup(name='collective.properties',
      version=version,
      description="Provides form to update object properties via Plone UI.",
      long_description=open("README.txt").read() + "\n" +
          open(os.path.join("collective", "properties", "TODO.txt")).read(
              ) + "\n" +
          open(os.path.join("docs", "HISTORY.txt")).read(),
      # Get more strings from
      # http://pypi.python.org/pypi?:action=list_classifiers
      classifiers=[
        "Programming Language :: Python",
        ],
      keywords='plone properties form ui',
      author='Vitaliy Podoba',
      author_email='vitaliypodoba@gmail.com',
      url='http://svn.plone.org/svn/collective/collective.properties',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['collective'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'plone.app.z3cform',
      ],
      entry_points="""
      # -*- entry_points -*-
      [z3c.autoinclude.plugin]
      target = plone
      """,
      )



from Acquisition import aq_inner

from zope.interface import implements
from zope.app.schema.vocabulary import IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary

from Products.CMFPlone.utils import normalizeString, safe_unicode


class SelectionPropertyTypeVocabulary(object):
    """Returns list of items returned by select_variable
    collable set into selection property type.
    """

    implements(IVocabularyFactory)

    def __call__(self, context):
      """Here we get Selection Property Type as context with set
      select_variable to call for list of items in vocabulary.
      """
      select_variable = context.select_variable
      context = aq_inner(context.context)
      if select_variable and hasattr(context, select_variable):
          values = getattr(context, select_variable)
          if callable(values):
              try:
                  values = values()
              except Exception:
                  return SimpleVocabulary([])
      else:
          return SimpleVocabulary([])
          
      # make sure our vocabulary returns only unicode strings so that
      # selection widget doesn't break with decoding error during render
      result = []
      for v in values:
          if isinstance(v, str):
              v = v.decode('utf-8')
          elif not isinstance(v, str):
              v = safe_unicode(str(v))
          
          # we also need to strip out every term in order to be compliant with
          # IPropertyManager interface
          v = v.strip()
          
          # we use custom widget so this token is not use in template,
          # only in widget converter
          token = normalizeString(v, encoding='utf-8')
          
          result.append(SimpleTerm(v, token, v))
      
      return SimpleVocabulary(result)



from zope.component import adapts
from zope.schema.interfaces import IBytesLine, IBytes, IChoice

from Products.CMFPlone.utils import safe_unicode, normalizeString

from z3c.form.interfaces import ITextAreaWidget, ITextWidget
from z3c.form.converter import BaseDataConverter, SequenceDataConverter

from collective.properties.interfaces import IBytesLineTuple, \
    ISpecialSelectWidget, ISpecialOrderedSelectWidget


class BytesLineTupleTextAreaDataConverter(BaseDataConverter):
    """A special converter between BytesLineTuple field and text area widget.
    
    Field is a tuple of utf-8 encoded strings while widget needs one utf-8
    encodded string containing above mentioned strings joined by line break.
    """

    adapts(IBytesLineTuple, ITextAreaWidget)

    def toWidgetValue(self, value):
        """Convert from Python sequence (tuple or list) to HTML representation:
        unicode string with line breaks between each sequence item.
        """
        if value is self.field.missing_value:
            return u''

        # we keep value in bytes field to be compliant with OFS.PropertyManager
        return u'\r\n'.join([safe_unicode(v) for v in value])

    def toFieldValue(self, value):
        """See interfaces.IDataConverter"""
        if value == u'':
            return self.field.missing_value
        
        collectionType = self.field._type
        if isinstance(collectionType, tuple):
            collectionType = collectionType[-1]
        
        return collectionType([v.encode('utf-8') for v in value.split(u'\r\n')])

class BytesLineTextDataConverter(BaseDataConverter):
    """A special converter between bytes line field and text widget."""

    adapts(IBytesLine, ITextWidget)

    def toWidgetValue(self, value):
        """Converts string to unicode"""
        if value is self.field.missing_value:
            return u''

        return safe_unicode(value)

    def toFieldValue(self, value):
        """See interfaces.IDataConverter"""
        if value == u'':
            return self.field.missing_value
        
        return value.encode('utf-8')

class BytesTextAreaDataConverter(BytesLineTextDataConverter):
    """A special converter between bytes field and text area widget."""

    adapts(IBytes, ITextAreaWidget)

class ChoiceSpecialSelectDataConverter(SequenceDataConverter):
    """A special converter between choice field and special select widget.
    It won't break if current field values is not present inside widget terms.
    
    Here we also are not using term token, but only value. This widget we adapt
    here works only with term value and title. Token is not used as
    it won't work for terms containing non-ascii characters in it's value and at
    the same time to stay compliant with IPropertyManager form.
    """
    adapts(IChoice, ISpecialSelectWidget)

    def toWidgetValue(self, value):
        widget = self.widget
        # if the value is the missing value, then an empty list is produced.
        if value is self.field.missing_value:
            return []
        
        # Look up the term in the terms
        terms = widget.updateTerms()
        if value not in terms:
            return []
        
        # ensure we get unicode for select widget, otherwise it breaks
        return [value]

    def toFieldValue(self, value):
        widget = self.widget
        if not len(value) or value[0] == widget.noValueToken:
            return self.field.missing_value
        
        # as we use widget that only uses term values this means
        # we got ready to use value and ensure we value is within
        # vocabulary terms
        terms = widget.updateTerms()
        if value[0] not in terms:
            return self.field.missing_value
        
        return value[0]

class BytesLineTupleSpecialOrderedSelectDataConverter(BaseDataConverter):
    """A special converter between bytesline tuple field and special ordered
    select widget.
    We use it for the same purposes as above defined converter for choice field.
    """

    adapts(IBytesLineTuple, ISpecialOrderedSelectWidget)

    def toWidgetValue(self, value):
        """Convert from Python bool to HTML representation."""
        widget = self.widget
        if widget.terms is None:
            widget.updateTerms()
        terms = widget.terms
        return [entry for entry in value if entry in terms]

    def toFieldValue(self, value):
        """See interfaces.IDataConverter"""
        widget = self.widget
        if widget.terms is None:
            widget.updateTerms()
        
        collectionType = self.field._type
        if isinstance(collectionType, tuple):
            collectionType = collectionType[-1]
        
        terms = widget.terms
        return collectionType([entry for entry in value if entry in terms])
